Index: netinet/udp_var.h
===================================================================
--- netinet/udp_var.h	(revision 185928)
+++ netinet/udp_var.h	(working copy)
@@ -107,6 +107,10 @@
 void		 udp_input(struct mbuf *, int);
 struct inpcb	*udp_notify(struct inpcb *inp, int errno);
 int		 udp_shutdown(struct socket *so);
+
+
+typedef void(*udp_tunnel_function)(struct mbuf *, int off);
+int udp_set_kernel_tunneling(struct socket *so, udp_tunnel_function f);
 #endif
 
 #endif
Index: netinet/udp_usrreq.c
===================================================================
--- netinet/udp_usrreq.c	(revision 185928)
+++ netinet/udp_usrreq.c	(working copy)
@@ -563,6 +563,18 @@
 		INP_RUNLOCK(inp);
 		goto badunlocked;
 	}
+	if (inp->inp_ppcb) {
+		/* Engage the tunneling protocol
+	 	 * we must make sure all locks
+		 * are released when we call the
+		 * tunneling protocol.
+		 */
+        	udp_tunnel_function tunnel_func;
+        	tunnel_func = (udp_tunnel_function)inp->inp_ppcb;
+        	INP_RUNLOCK(inp);
+        	tunnel_func(m, iphlen);
+        	return;
+    	}
 	udp_append(inp, ip, m, iphlen + sizeof(struct udphdr), &udp_in);
 	INP_RUNLOCK(inp);
 	return;
@@ -1138,10 +1150,45 @@
 	INP_INFO_WUNLOCK(&V_udbinfo);
 	inp->inp_vflag |= INP_IPV4;
 	inp->inp_ip_ttl = V_ip_defttl;
+	/*
+	 * UDP does not have a per-protocol
+	 * pcb (inp->inp_ppcb). We use this
+	 * pointer for kernel tunneling pointer.
+	 * If we ever need to have a protocol
+	 * block we will need to move this
+	 * function pointer there. Null
+	 * in this pointer means "do the normal
+	 * thing".
+	 */
+	inp->inp_ppcb = NULL;
 	INP_WUNLOCK(inp);
 	return (0);
 }
 
+int
+udp_set_kernel_tunneling(struct socket *so, udp_tunnel_function f)
+{
+	struct inpcb *inp;
+	inp = (struct inpcb *)so->so_pcb;
+
+	if (so->so_type != SOCK_DGRAM) {
+		/* Not UDP socket... sorry */
+		return (ENOTSUP);
+	}
+	if (inp == NULL) {
+		/* NULL INP? */
+		return (EINVAL);
+	}
+	if (inp->inp_lport == 0) {
+		/* Not bound */
+		return (EINVAL);
+	}
+	INP_WLOCK(inp);
+	inp->inp_ppcb = f;
+	INP_WUNLOCK(inp);
+	return (0);
+}
+
 static int
 udp_bind(struct socket *so, struct sockaddr *nam, struct thread *td)
 {
Index: netinet6/udp6_usrreq.c
===================================================================
--- netinet6/udp6_usrreq.c	(revision 185928)
+++ netinet6/udp6_usrreq.c	(working copy)
@@ -354,6 +354,18 @@
 	}
 	INP_RLOCK(inp);
 	INP_INFO_RUNLOCK(&V_udbinfo);
+	if (inp->inp_ppcb) {
+		/* Engage the tunneling protocol
+		 * we must make sure all locks
+		 * are released when we call the
+		 * tunneling protocol.
+		 */
+		udp_tunnel_function tunnel_func;
+		tunnel_func = (udp_tunnel_function)inp->inp_ppcb;
+		INP_RUNLOCK(inp);
+		tunnel_func(m, off);
+		return (IPPROTO_DONE);
+	}
 	udp6_append(inp, m, off, &fromsa);
 	INP_RUNLOCK(inp);
 	return (IPPROTO_DONE);
